@@ -99,6 +99,57 @@ class IArchive(Interface):
         """Returns the contents of a container as IContainerRecord.
         """
 
+    def iter_hierarchy(top_container_id, max_depth=None,
+            follow_deleted=False, follow_moved=False):
+        """Iterate over IContainerRecords in a hierarchy.
+
+        This is more efficient than traversing the hierarchy
+        by calling the container_contents method repeatedly, because
+        this method minimizes the number of calls to the database.
+        Yields an IContainerRecord for each container in the
+        hierarchy.
+
+        Set the max_depth parameter to limit the depth of containers
+        to include. When max_depth is 0, only the top container is included;
+        when depth is 1, its children are included, and so on.
+        Set max_depth to None (the default) to include containers
+        of arbitrary depth.
+
+        If follow_deleted is true, this method will also include
+        deleted containers and descendants of deleted containers
+        in the results.
+
+        If follow_moved is true, this method will also include
+        moved containers and descendants of moved containers
+        in the results.
+
+        If no container exists by the given top_container_id, this
+        method returns an empty mapping.
+
+        NB: This method assumes that container_ids are also docids.
+        (Most other methods make no such assumption.)
+        """
+
+    def filter_container_ids(container_ids):
+        """Returns which of the specified container IDs exist in the archive.
+
+        Returns a sequence containing a subset of the provided container_ids.
+        """
+
+    def which_contain_deleted(self, container_ids, max_depth=None):
+        """Returns the subset of container_ids that have had something deleted.
+
+        This is useful for building a hierarchical trash UI that allows
+        users to visit only containers that have had something deleted.
+        All descendant containers are examined (unless max_depth limits
+        the search.)
+
+        Returns a sequence containing a subset of the provided container_ids.
+
+        NB: This method assumes that container_ids are also docids.
+        (Most other methods make no such assumption.)
+        """
+
 
 class IObjectVersion(IDCDescriptiveProperties, IDCTimes):
     """The content of an object for version control.
@@ -237,3 +288,8 @@ class IDeletedItem(Interface):
 
     Empty or None if the object is not currently in any container.
     """)
+
+    moved = Attribute("""True if this item was moved rather than deleted.
+
+    True when new_container_ids is non-empty.
+    """)
