@@ -31,8 +31,8 @@ class ArchiveTest(unittest.TestCase):
         params = EngineParams('sqlite:///')
         return self._make(params)
 
-    def _make_dummy_object_version(self):
-        return DummyObjectVersion()
+    def _make_dummy_object_version(self, docid=4):
+        return DummyObjectVersion(docid)
 
     def test_verifyImplements_IArchive(self):
         from zope.interface.verify import verifyClass
@@ -369,6 +369,7 @@ class ArchiveTest(unittest.TestCase):
         self.assertTrue(records[0].blobs)
         self.assertEqual(records[0].blobs.keys(), ['x'])
         blob = records[0].blobs['x']
+        self.assertEqual(blob.getvalue(), '42')  # StringIO has getvalue()
         self.assertEqual(blob.read(), '42')
         with self.assertRaises(IOError):
             blob.write('x')
@@ -377,6 +378,33 @@ class ArchiveTest(unittest.TestCase):
 
         self.assertFalse(records[1].blobs)
 
+    def test_history_with_tempfile_blob(self):
+        obj = self._make_dummy_object_version()
+        archive = self._make_default()
+        archive.archive(obj)
+        obj.blobs = {'x': StringIO('data' * 1000)}
+        archive.archive(obj)
+
+        from repozitory.schema import ArchivedChunk
+        rows = archive.session.query(ArchivedChunk).all()
+        self.assertEqual(len(rows), 1)
+
+        records = archive.history(obj.docid)
+        self.assertEqual(len(records), 2)
+
+        self.assertTrue(records[0].blobs)
+        self.assertEqual(records[0].blobs.keys(), ['x'])
+        blob = records[0].blobs['x']
+        blob._max_stringio = 100
+        self.assertFalse(hasattr(blob, 'getvalue'))
+        self.assertEqual(blob.read(), 'data' * 1000)
+        with self.assertRaises(IOError):
+            blob.write('x')
+        with self.assertRaises(IOError):
+            blob.writelines(['x'])
+
+        self.assertFalse(records[1].blobs)
+
     def test_history_with_blob_having_multiple_chunks(self):
         archive = self._make_default()
         archive.chunk_size = 13
@@ -896,9 +924,230 @@ class ArchiveTest(unittest.TestCase):
         self.assertTrue(r.deleted[0].deleted_time)
         self.assertEqual(r.deleted[0].new_container_ids, [6])
 
+    def _make_hierarchy(self, archive, delete_c6=True, move_c7=False):
+
+        class DummyContainerVersion:
+            def __init__(self, container_id, path):
+                self.container_id = container_id
+                self.path = path
+                self.map = {}
+                self.ns_map = {}
+
+        # Create a hierarchy:
+        #
+        # c4
+        #   - c5
+        #   - c6 (optionally deleted by user2)
+        #     - c7 (deleted by user2)
+        #     - c8
+        # c9
+        #   - c7 (optionally moved by user3)
+
+        c8 = DummyContainerVersion(8, '/c4/c6/c8')
+        archive.archive_container(c8, 'user1')
+
+        c7 = DummyContainerVersion(7, '/c4/c6/c7')
+        archive.archive_container(c7, 'user1')
+
+        c6 = DummyContainerVersion(6, '/c4/c6')
+        c6.map = {'c7': 7, 'c8': 8}
+        archive.archive_container(c6, 'user1')
+
+        c5 = DummyContainerVersion(5, '/c4/c5')
+        archive.archive_container(c5, 'user1')
+
+        c4 = DummyContainerVersion(4, '/c4')
+        c4.map = {'c5': 5, 'c6': 6}
+        archive.archive_container(c4, 'user1')
+
+        del c6.map['c7']
+        archive.archive_container(c6, 'user2')
+        if delete_c6:
+            del c4.map['c6']
+            archive.archive_container(c4, 'user2')
+
+        if move_c7:
+            c9 = DummyContainerVersion(9, '/c9')
+            c9.map = {'c7': 7}
+            archive.archive_container(c9, 'user3')
+
+        return c4
+
+    def test_iter_hierarchy_with_max_depth_0(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive)
+
+        containers = dict((record.container_id, record) for record in
+            archive.iter_hierarchy(4, max_depth=0, follow_deleted=True))
+        self.assertEqual(set(containers.keys()), set([4]))
+
+    def test_iter_hierarchy_with_max_depth_1(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive)
+
+        containers = dict((record.container_id, record) for record in
+            archive.iter_hierarchy(4, max_depth=1, follow_deleted=True))
+        self.assertEqual(set(containers.keys()), set([4, 5, 6]))
+
+    def test_iter_hierarchy_with_max_depth_2(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive)
+
+        containers = dict((record.container_id, record) for record in
+            archive.iter_hierarchy(4, max_depth=2, follow_deleted=True))
+        self.assertEqual(set(containers.keys()), set([4, 5, 6, 7, 8]))
+
+    def test_iter_hierarchy_without_follow_deleted(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive)
+
+        containers = dict((record.container_id, record) for record in
+            archive.iter_hierarchy(4))
+        self.assertEqual(set(containers.keys()), set([4, 5]))
+
+        r = containers[4]
+        self.assertEqual(r.map, {'c5': 5})
+        self.assertEqual(r.ns_map, {})
+        self.assertEqual(len(r.deleted), 1)
+
+        row = r.deleted[0]
+        from zope.interface.verify import verifyObject
+        from repozitory.interfaces import IDeletedItem
+        verifyObject(IDeletedItem, row)
+
+        self.assertEqual(r.deleted[0].docid, 6)
+        self.assertEqual(r.deleted[0].namespace, '')
+        self.assertEqual(r.deleted[0].name, 'c6')
+        self.assertEqual(r.deleted[0].deleted_by, 'user2')
+        self.assertTrue(r.deleted[0].deleted_time)
+        self.assertEqual(r.deleted[0].new_container_ids, None)
+
+        r = containers[5]
+        self.assertEqual(r.map, {})
+        self.assertEqual(r.ns_map, {})
+        self.assertEqual(len(r.deleted), 0)
+
+    def test_iter_hierarchy_with_follow_deleted_and_no_moves(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive)
+
+        containers = dict((record.container_id, record) for record in
+            archive.iter_hierarchy(4, follow_deleted=True))
+        self.assertEqual(set(containers.keys()), set([4, 5, 6, 7, 8]))
+
+        r = containers[6]
+        self.assertEqual(r.map, {'c8': 8})
+        self.assertEqual(r.ns_map, {})
+        self.assertEqual(len(r.deleted), 1)
+
+        row = r.deleted[0]
+        from zope.interface.verify import verifyObject
+        from repozitory.interfaces import IDeletedItem
+        verifyObject(IDeletedItem, row)
+
+        self.assertEqual(r.deleted[0].docid, 7)
+        self.assertEqual(r.deleted[0].namespace, '')
+        self.assertEqual(r.deleted[0].name, 'c7')
+        self.assertEqual(r.deleted[0].deleted_by, 'user2')
+        self.assertTrue(r.deleted[0].deleted_time)
+        self.assertEqual(r.deleted[0].new_container_ids, None)
+
+        r = containers[7]
+        self.assertEqual(r.map, {})
+        self.assertEqual(r.ns_map, {})
+        self.assertEqual(len(r.deleted), 0)
+
+    def test_iter_hierarchy_with_follow_deleted_and_1_move(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive, move_c7=True)
+
+        containers = dict((record.container_id, record) for record in
+            archive.iter_hierarchy(4, follow_deleted=True))
+        self.assertEqual(set(containers.keys()), set([4, 5, 6, 8]))
+
+        r = containers[6]
+        self.assertEqual(r.map, {'c8': 8})
+        self.assertEqual(r.ns_map, {})
+        self.assertEqual(len(r.deleted), 1)
+
+        row = r.deleted[0]
+        from zope.interface.verify import verifyObject
+        from repozitory.interfaces import IDeletedItem
+        verifyObject(IDeletedItem, row)
+
+        self.assertEqual(r.deleted[0].docid, 7)
+        self.assertEqual(r.deleted[0].namespace, '')
+        self.assertEqual(r.deleted[0].name, 'c7')
+        self.assertEqual(r.deleted[0].deleted_by, 'user2')
+        self.assertTrue(r.deleted[0].deleted_time)
+        self.assertEqual(r.deleted[0].new_container_ids, [9])
+
+    def test_iter_hierarchy_with_follow_deleted_and_follow_moved(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive, move_c7=True)
+
+        containers = dict((record.container_id, record) for record in
+            archive.iter_hierarchy(4, follow_deleted=True, follow_moved=True))
+        self.assertEqual(set(containers.keys()), set([4, 5, 6, 7, 8]))
+
+        r = containers[6]
+        self.assertEqual(r.map, {'c8': 8})
+        self.assertEqual(r.ns_map, {})
+        self.assertEqual(len(r.deleted), 1)
+
+        row = r.deleted[0]
+        from zope.interface.verify import verifyObject
+        from repozitory.interfaces import IDeletedItem
+        verifyObject(IDeletedItem, row)
+
+        self.assertEqual(r.deleted[0].docid, 7)
+        self.assertEqual(r.deleted[0].namespace, '')
+        self.assertEqual(r.deleted[0].name, 'c7')
+        self.assertEqual(r.deleted[0].deleted_by, 'user2')
+        self.assertTrue(r.deleted[0].deleted_time)
+        self.assertEqual(r.deleted[0].new_container_ids, [9])
+
+    def test_filter_container_ids(self):
+        archive = self._make_default()
+        c4 = self._make_hierarchy(archive)
+        obj3 = self._make_dummy_object_version(3)
+        archive.archive(obj3)
+        c4.map['obj3'] = 3
+        archive.archive_container(c4, 'user4')
+        actual = archive.filter_container_ids([3, 4, 5])
+        expect = [4, 5]
+        self.assertEqual(set(expect), set(actual))
+
+    def test_which_contain_deleted_simple(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive)
+        actual = archive.which_contain_deleted([5, 6])
+        expect = [6]
+        self.assertEqual(set(expect), set(actual))
+
+    def test_which_contain_deleted_with_successful_depth_traversal(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive, delete_c6=False)
+        actual = archive.which_contain_deleted([4])
+        expect = [4]
+        self.assertEqual(set(expect), set(actual))
+
+    def test_which_contain_deleted_with_traversal_exceeding_max_depth(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive, delete_c6=False)
+        actual = archive.which_contain_deleted([4], max_depth=0)
+        expect = []
+        self.assertEqual(set(expect), set(actual))
+
+    def test_which_contain_deleted_ignore_moved(self):
+        archive = self._make_default()
+        self._make_hierarchy(archive, delete_c6=False, move_c7=True)
+        actual = archive.which_contain_deleted([5, 6])
+        expect = []
+        self.assertEqual(set(expect), set(actual))
+
 
 class DummyObjectVersion:
-    docid = 4
     path = '/my/object'
     created = datetime.datetime(2011, 4, 6)
     modified = datetime.datetime(2011, 4, 7)
@@ -907,3 +1156,6 @@ class DummyObjectVersion:
     attrs = {'a': 1, 'b': [2]}
     user = 'tester'
     comment = 'I like version control.'
+
+    def __init__(self, docid):
+        self.docid = docid
